#include <petscdmda.h>
#include <petscksp.h>
//#include <petscpcmg.h>
#include <mpi.h>
#include "ComputeFunctions.h"
#include "GlobalVariables.h"

// -- 0.0 --  "ComputerFunctions" Global Functions
extern PetscErrorCode BuildNodes();


#undef __FUNCT__
#define __FUNCT__ "ApplyNonUniformGrid"
PetscErrorCode ApplyNonUniformGrid(DM da) {
  DM             cda;
  Vec            coords;
  DMDACoor3d       ***_coords;
  PetscInt       si,sj,sk,nx,ny,nz;
  PetscErrorCode ierr;

  PetscFunctionBeginUser;

  ierr = DMGetCoordinateDM(da,&cda);CHKERRQ(ierr);
  ierr = DMGetCoordinatesLocal(da,&coords);CHKERRQ(ierr);
  ierr = DMDAVecGetArray(cda,coords,&_coords);CHKERRQ(ierr);
  ierr = DMDAGetGhostCorners(cda,&si,&sj,&sk,&nx,&ny,&nz);CHKERRQ(ierr);
//  printf("%d  %d  %d  %d  %d  %d \n", si, sj, sk, nx, ny, nz);

  for (k = sk; k < sk+nz; k++) {
    for (j = sj; j < sj+ny; j++) {
      for (i = si; i < si+nx; i++) {
        PetscRealPart(_coords[k][j][i].x) = PhyX[i];
        PetscRealPart(_coords[k][j][i].y) = PhyY[j];
        PetscRealPart(_coords[k][j][i].z) = PhyZ[k];
      }
    }
  }

  ierr = DMDAVecRestoreArray(cda,coords,&_coords);CHKERRQ(ierr);

  PetscFunctionReturn(0);
}


#undef __FUNCT__
#define __FUNCT__ "BuildNonUniformGrid"

PetscErrorCode BuildNonUniformGrid() {
  double dxValue, dyValue, dzValue;

  dxValue = dx;
  dyValue = dy;
  dzValue = dz;

  PetscFunctionBegin;
  
  // Effective mesh is from i = 0~imax-1
  // need i = imax data for PhyDx for generate uDx later
  PhyX[0] = 0.;
  for(i = 1; i <= imax; i++) {
    if(i >= 1 && i <= seci1) {
      PhyDx[i] = dxValue*pow(rdx1,(seci1-i));
    } else if(i > seci1 && i <= seci1+seci2) {
      PhyDx[i] = dxValue;
    } else if(i > seci1+seci2 && i <= seci1+seci2+seci3) {
      PhyDx[i] = dxValue*pow(rdx3,(i-seci1-seci2));
    } else { printf("ERROR for generating non-uniform mesh, i = %d!!!!\n", i); }
    PhyX[i] = PhyX[i-1] + PhyDx[i];
  }
  PhyDx[0] = PhyDx[1];
  GLeftDx = PhyDx[0];
  GRightDx = PhyDx[imax-1];
  // i = 0 is the computational points, this is also good for generating mesh later. 
  //         0       1       2        3
  //  ...____|_______|_______|_______|___...
  //           dx(1)   dx(2)   dx(3)
  for(i = 0; i <= xb+xg; i++) {
    GLeftX[i] = PhyX[0] - GLeftDx*i;
    GRightX[i] = PhyX[imax-1] + GRightDx*i; 
  }

  PhyY[0] = 0.;
  for(j = 1; j <= jmax; j++) {
    if(j >= 1 && j <= secj1) {
      PhyDy[j] = dyValue*pow(rdy1,(secj1-j));
    } else if(j > secj1 && j <= secj1+secj2) {
      PhyDy[j] = dyValue;
    } else if(j > secj1+secj2 && j <= secj1+secj2+secj3) {
      PhyDy[j] = dyValue*pow(rdy3,(j-secj1-secj2));
    } else { printf("ERROR for generating non-uniform mesh, j = %d!!!!\n", j); }
    PhyY[j] = PhyY[j-1] + PhyDy[j];
  }
  PhyDy[0] = PhyDy[1];
  GBotDy = PhyDy[0];
  GTopDy = PhyDy[jmax-1];
  for(j = 0; j <= yb+yg; j++) {
    GBotY[j] = PhyY[0] - GBotDy*j;
    GTopY[j] = PhyY[jmax-1] + GTopDy*j;
  }

  PhyZ[0] = 0.;
  for(k = 1; k <= kmax; k++) {
    if(k >= 1 && k <= seck1) {
      PhyDz[k] = dzValue*pow(rdz1,(seck1-k));
    } else if(k > seck1 && k <= seck1+seck2) {
      PhyDz[k] = dzValue;
    } else if(k > seck1+seck2 && k <= seck1+seck2+seck3) {
      PhyDz[k] = dzValue*pow(rdz3,(k-seck1-seck2));
    } else { printf("ERROR for generating non-uniform mesh, k = %d!!!!\n", k); }
    PhyZ[k] = PhyZ[k-1] + PhyDz[k];
  }
  PhyDz[0] = PhyDz[1];
  GAftDz = PhyDz[0];
  GFrontDz = PhyDz[kmax-1];
  for(k = 0; k <= zb+zg; k++) {
    GAftZ[k] = PhyZ[0] - GAftDz*k;
    GFrontZ[k] = PhyZ[kmax-1] + GFrontDz*k;
  }


  PetscFunctionReturn(0);

}



#undef __FUNCT__
#define __FUNCT__ "Initialization"

PetscErrorCode Initialization(DM da, int time) {
  MPI_Status status;
  DM             cda;
  PetscErrorCode ierr;
  Vec gc;
  FILE *fppi;

  PetscFunctionBegin;
// -- 1.0 -- Read Info. of DMDA
  ierr = DMDAGetInfo(da, 0, &mx, &my, &mz, &mm , &nn, &pp,0,0,0,0,0,0);CHKERRQ(ierr);
  MPI_Comm_rank(PETSC_COMM_WORLD, &rank);
  ierr = DMDAGetCorners(da,&xs,&ys,&zs,&xw,&yw,&zw);CHKERRQ(ierr);

  ierr = DMGetCoordinateDM(da,&cda);
  ierr = DMGetCoordinatesLocal(da,&gc);
  ierr = DMDAVecGetArray(cda, gc, &coors);


    ierr = BuildNodes();


  ierr = DMDAVecRestoreArray(cda, gc, &coors);
  
  PetscFunctionReturn(0);

}


#undef __FUNCT__
#define __FUNCT__ "BuildNodes"

PetscErrorCode BuildNodes() {
  int counti, countj, countk;  

  PetscFunctionBegin;

  // -- 1.0 -- Determine the size of local computational grid
  xc = xw;
  yc = yw;
  zc = zw;
    // if it's Neumann BC, you do not need to reduce the computational grid size, i.e.
    // for a 10*10 grid with IB pressure BC's: left, bot, top = Neumann, and right = Dirichlet
    // left: (i=0) == (i=-1); right: (i=imax-1) = contant;

  NbrRank.l = rank-1;
  NbrRank.r = rank+1;
  NbrRank.b = rank-mm;
  NbrRank.t = rank+mm;
  NbrRank.a = rank-mm*nn;
  NbrRank.f = rank+mm*nn; 

  uDx = (double*) calloc((xc+xg*2+xb*2), sizeof(double));
  vDx = (double*) calloc((xc+xg*2+xb*2), sizeof(double));
  wDx = (double*) calloc((xc+xg*2+xb*2), sizeof(double));
  pDx = (double*) calloc((xc+xg*2+xb*2), sizeof(double));

  uDy = (double*) calloc((yc+yg*2+yb*2), sizeof(double));
  vDy = (double*) calloc((yc+yg*2+yb*2), sizeof(double));
  wDy = (double*) calloc((yc+yg*2+yb*2), sizeof(double));
  pDy = (double*) calloc((yc+yg*2+yb*2), sizeof(double));

  uDz = (double*) calloc((zc+zg*2+zb*2), sizeof(double));
  vDz = (double*) calloc((zc+zg*2+zb*2), sizeof(double));
  wDz = (double*) calloc((zc+zg*2+zb*2), sizeof(double));
  pDz = (double*) calloc((zc+zg*2+zb*2), sizeof(double));


  
  // -- 3.0 -- Fill up local indices
  // in x-direction
  // 0 ~ xg-1 --> left ghosted nodes;                   xg+2*xb+xc ~ xg*2+xb*2+xc --> right ghosted nodes
  _Index.lgEnd = xg-1; // left ghost ending pt
  _Index.rgStr = xg+2*xb+xc; // right ghost starting pt
  // xg ~ xg+xb-1 --> left boundary nodes;              xg+xb+xc ~ xg+2*xb+xc-1 --> right boundary nodes
  _Index.lbEnd = xg+xb-1;
  _Index.rbStr = xg+xb+xc;
  // xg+xb ~ xg+xb+xc-1 --> computational nodes; 
  _Index.lcStr = xg+xb;
  _Index.rcEnd = xg+xb+xc-1;

  // 0 ~ yg-1 --> bottom ghosted nodes;                   yg+2*yb+yc ~ yg*2+yb*2+yc --> top ghosted nodes
  _Index.bgEnd = yg-1;
  _Index.tgStr = yg+2*yb+yc;
  // yg ~ yg+yb-1 --> bottom boundary nodes;              yg+yb+yc ~ yg+2*yb+yc-1 --> top boundary nodes
  _Index.bbEnd = yg+yb-1;
  _Index.tbStr = yg+yb+yc;
  // yg+yb ~ yg+yb+yc-1 --> computational nodes; 
  _Index.bcStr = yg+yb;
  _Index.tcEnd = yg+yb+yc-1;

  // 0 ~ zg-1 --> bottom ghosted nodes;                   zg+2*zb+zc ~ zg*2+zb*2+zc --> top ghosted nodes
  _Index.agEnd = zg-1;
  _Index.fgStr = zg+2*zb+zc;
  // zg ~ zg+zb-1 --> bottom boundary nodes;              zg+zb+zc ~ zg+2*zb+zc-1 --> top boundary nodes
  _Index.abEnd = zg+zb-1;
  _Index.fbStr = zg+zb+zc;
  // zg+zb ~ zg+zb+zc-1 --> computational nodes; 
  _Index.acStr = zg+zb;
  _Index.fcEnd = zg+zb+zc-1;

  // -- 3.0 -- Finished
  double Test1, Test2, Test3;
  double HybridX, HybridY, HybridZ;
  double HybridDx, HybridDy, HybridDz;
  double HybridDx1, HybridDy1, HybridDz1;
  for(i = 0; i < xc+xg*2+xb*2; i++) {
    for(j = 0; j < yc+yg*2+yb*2; j++) {
      for(k = 0; k < zc+zg*2+zb*2; k++) {
        counti = i-_Index.lcStr+xs;  // coodiante of coors = coodinate of Alan's grid - _Index.lcStr + xs , therefore, see 'DetermineFlags'
        countj = j-_Index.bcStr+ys;
        countk = k-_Index.acStr+zs;

        if(i < _Index.lcStr && localBC.lBC) {
          HybridX = GLeftX[_Index.lcStr-i];
          HybridDx = GLeftDx;
          HybridDx1 = GLeftDx;
        } else if(i > _Index.rcEnd+localBC.rBC && localBC.rBC) {
          HybridX = GRightX[i-_Index.rcEnd-localBC.rBC];
          HybridDx = GRightDx;
          HybridDx1 = GRightDx;
        } else {
          HybridX = PhyX[counti];
          HybridDx = PhyDx[counti];
          HybridDx1  = (i==_Index.lcStr&&localBC.lBC)?GLeftDx:PhyDx[counti-1];
        }

        if(j < _Index.bcStr && localBC.bBC) {
          HybridY = GBotY[_Index.bcStr-j];
          HybridDy = GBotDy;
          HybridDy1 = GBotDy;
        } else if(j > _Index.tcEnd && localBC.tBC) {
          HybridY = GTopY[j-_Index.tcEnd];
          HybridDy = GTopDy;
          HybridDy1 = GTopDy;
        } else {
          HybridY = PhyY[countj];
          HybridDy = PhyDy[countj];
          HybridDy1 = (j==_Index.bcStr&&localBC.bBC)?GBotDy:PhyDy[countj-1];
        }

        if(k < _Index.acStr && localBC.aBC) {
          HybridZ = GAftZ[_Index.acStr-k];
          HybridDz = GAftDz;
          HybridDz1 = GAftDz;
        } else if(k > _Index.fcEnd && localBC.fBC) {
          HybridZ = GFrontZ[k-_Index.fcEnd];
          HybridDz = GFrontDz;
          HybridDz1 = GFrontDz;
        } else {
          HybridZ = PhyZ[countk];
          HybridDz = PhyDz[countk];
          HybridDz1 = (k==_Index.acStr&&localBC.aBC)?GAftDz:PhyDz[countk-1];
        }

          uDx[i] = HybridDx;
          uDy[j] = (HybridDy+HybridDy1)/2.;
          uDz[k] = (HybridDz+HybridDz1)/2.;

          vDx[i] = (HybridDx+HybridDx1)/2.;
          vDy[j] = HybridDy;
          vDz[k] = (HybridDz+HybridDz1)/2.;

          wDx[i] = (HybridDx+HybridDx1)/2.;
          wDy[j] = (HybridDy+HybridDy1)/2.;
          wDz[k] = HybridDz;

          pDx[i] = (HybridDx+HybridDx1)/2.;
          pDy[j] = (HybridDy+HybridDy1)/2.;
          pDz[k] = (HybridDz+HybridDz1)/2.;
	  
      }
    }
  }

  // -- 4.0 -- Finished

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ComputePpRHS"

PetscErrorCode ComputePpRHS(KSP ksp, Vec b, void *ctx) {
  PetscErrorCode ierr;
  DM da;
  PetscScalar    ***array;
  double firstT, secondT;
  int arrayi, arrayj, arrayk;
  double dxValue, dyValue, dzValue;


  PetscFunctionBegin;
  KSPGetDM(ksp,&da);
  ierr = DMDAGetInfo(da,0,&mx,&my,&mz,0,0,0,0,0,0,0,0,0);CHKERRQ(ierr); 

  ierr = DMDAVecGetArray(da, b, &array);CHKERRQ(ierr);

  if((_Index.rcEnd+localBC.rBC-_Index.lcStr+1) != xw ) printf("Rank#%d, X-direction width ERROR!!!!\n", rank);
  if((_Index.tcEnd-_Index.bcStr+1) != yw ) printf("Rank#%d, Y-direction width ERROR!!!!\n", rank);
  if((_Index.fcEnd-_Index.acStr+1) != zw ) printf("Rank#%d, Z-direction width ERROR!!!!\n", rank);
  for(i = _Index.lcStr; i <= _Index.rcEnd+localBC.rBC; i++) {
    for(j = _Index.bcStr; j <= _Index.tcEnd; j++) {
      for(k = _Index.acStr; k <= _Index.fcEnd; k++) {
	
	arrayi = i + xs - _Index.lcStr; // coodinate of coors = coodiante of Alan's grid + xs - _Index.lcStr 
	arrayj = j + ys - _Index.bcStr;
	arrayk = k + zs - _Index.acStr;
        dxValue = uDx[i];
        dyValue = vDy[j];
        dzValue = wDz[k];
     
//	array[arrayk][arrayj][arrayi] = 0.0;
        array[arrayk][arrayj][arrayi] = -1.*sin(2.*PI*i/xw)*sin(4.*PI*j/yw)*sin(6.*PI*k/zw)*dxValue*dyValue*dzValue;
      }
    }
  }

  ierr = DMDAVecRestoreArray(da, b, &array);CHKERRQ(ierr);
  ierr = VecAssemblyBegin(b);CHKERRQ(ierr);
  ierr = VecAssemblyEnd(b);CHKERRQ(ierr);

  // force right hand side to be consistent for singular matrix 
  // note this is really a hack, normally the model would provide you with a consistent right handside


  PetscFunctionReturn(0);

}

