// grid dimension is : [((imax-1)*4+1) * ((jmax-1)*4+1)] with ghost grids and [((imax-1)*4-1) * ((jmax-1)*4-1)] without ghost grids
// domain dimension is : [{-dxValue ~ ((imax-1)*4-1)*dxValue} * {-dyValue ~ ((jmax-1)*4-1)*dyValue}]  with ghost grids 
//[{0 ~ ((imax-1)*4-2)*dxValue} * {0 ~ ((jmax-1)*4-2)*dyValue}]  without ghost grids

#define cnti 110 
#define cntj 100 
#define cntk 100 
#define imax 300 
#define jmax 200
#define kmax 200
#define dx 0.01
#define dy 0.01
#define dz 0.01

#define seci1 60 
#define seci2 100
#define seci3 140
#define secj1 50
#define secj2 100
#define secj3 50
#define seck1 50
#define seck2 100
#define seck3 50
/*
#define cnti 80
#define cntj 48 
#define cntk 48
#define imax 200 
#define jmax 96
#define kmax 96
#define dx 0.025
#define dy 0.025
#define dz 0.025

#define seci1 60
#define seci2 40
#define seci3 100
#define secj1 28
#define secj2 40
#define secj3 28
#define seck1 28
#define seck2 40
#define seck3 28
*/
#define rdx1 1.07
#define rdx3 1.03
#define rdy1 1.1
#define rdy3 1.1
#define rdz1 1.1
#define rdz3 1.1
/*
#define rdx1 1.
#define rdx3 1.
#define rdy1 1.
#define rdy3 1.
#define rdz1 1.
#define rdz3 1.
*/
#define MaxGridShift 2 // maximum grid shift in one time step // THIS MAY CAUSE SOME PROBLEM IN CASE WITH PERIODIC BC, SINCE IT REQUIRES THE BOUNDARY NODES HAVING SOME DISTANCE FROM THE GEOMETRY

#define gistar seci1
#define gistop seci1+seci2
#define gjstar secj1
#define gjstop secj1+secj2
#define gkstar seck1
#define gkstop seck1+seck2

#define nmpcrcl 180
#define CD (nmpcrcl*nmpcrcl)
#define rr 0.5
#define PI 3.1415926535
#define inlet 1.0
#define mu (double)(1./300.) 
#define dt 0.001
#define outputInt 10

#define MirrorVgridBC 0 
#define xg 2 
#define yg 2
#define zg 2
#define xb 2
#define yb 2
#define zb 2

typedef enum {DIRICHLET, NEUMANN} BCType;

typedef struct {
  int l;
  int r;
  int t;
  int b;
  int f;
  int a; // aft = back
} VicinityGrid; // Grid in vicinity

typedef struct {
  double x[CD], y[CD], z[CD];
  double x2[CD], y2[CD], z2[CD];
  int Index[CD], Index2[CD]; // record index numbers of geometry within local grid range
  int Count, Count2;
//  int YesGeo; // Determine if local grids are immersed inside the geometry
  VicinityGrid vg;
  double bfu[CD], bfv[CD], bfw[CD]; // used in boundary force calculations
} Geometry;

typedef struct {
  int uc;
  int vc;
  int wc;
  int pc;
} Flag;

typedef struct {
  double ffu;
  double ffv;
  double ffw;
  double ffuo;
  double ffvo;
  double ffwo;
} ForcingTerm;

typedef struct {
  double x;
  double y;
  double z;
} ThreeDGrid;

typedef struct {
  ThreeDGrid ucGrid;
  ThreeDGrid vcGrid;
  ThreeDGrid wcGrid;
  ThreeDGrid pcGrid;
  Flag fg;
  ForcingTerm ft;
  VicinityGrid vg;  
} StaggerGrid;

typedef struct {
  double u;  double v;  double w; double p;
  double uo;  double vo;  double wo; double po;
  double uoo;  double voo;  double woo; double dpo;
  double duudx;  double duvdy;  double duwdz;
  double dvudx;  double dvvdy;  double dvwdz;
  double dwudx;  double dwvdy;  double dwwdz;
  double duudxo;  double duvdyo;  double duwdzo;
  double dvudxo;  double dvvdyo;  double dvwdzo;
  double dwudxo;  double dwvdyo;  double dwwdzo;
  double dp;
} NodeV;

typedef struct {
  double u;  double v;  double w; double p;
  double dp;
} NodePrimV;

typedef struct {
  // x-direction
  int lgEnd; // the max number of left ghosted nodes
  int rgStr; // the min number of right ghosted nodes
  int lbEnd; // the max number of left boundary nodes
  int rbStr; // the min number of right boundary nodes
  int lcStr; // the min number of computational nodes
  int rcEnd; // the max number of computational nodes

  // y-direction
  int bgEnd; // the max number of bottom ghosted nodes
  int tgStr; // the min number of top ghosted nodes
  int bbEnd; // the max number of bottom boundary nodes
  int tbStr; // the min number of top boundary nodes
  int bcStr; // the min number of computational nodes
  int tcEnd; // the max number of computational nodes

  // z-direction
  int agEnd; // the max number of bottom ghosted nodes
  int fgStr; // the min number of top ghosted nodes
  int abEnd; // the max number of bottom boundary nodes
  int fbStr; // the min number of top boundary nodes
  int acStr; // the min number of computational nodes
  int fcEnd; // the max number of computational nodes

} NodeIndex;

typedef struct {
  int lBC;
  int rBC;
  int bBC;
  int tBC;
  int aBC;
  int fBC;

} LocalBC;

typedef struct {
  int l;
  int r;
  int b;
  int t;
  int a;
  int f;
} NeighborRank;

typedef struct {
  int L1st;
  int L2nd;
  int R1st;
  int R2nd;
  int T1st;
  int T2nd;
  int B1st;
  int B2nd;
  int F1st;
  int F2nd;
  int A1st;
  int A2nd;
} CommunicationTag;
